#include "mainwindow.h"
#include "qpunchcard/card.h"
#include "qpunchcard/widget.h"
#include "deckviewer/navigatormodel.h"

#include "qpunchcard/jones.h"

#include <QApplication>
#include <QLabel>
#include <QtDebug>
#include <QListView>
#include <QUndoView>
#include <QDockWidget>
#include <QHBoxLayout>
#include <QPushButton>
#include <QMessageBox>
#include <QFileDialog>
#include <QFileInfo>

using namespace QPunchCard;
using namespace App;

MainWindow::MainWindow() {
	// sehr praktisch:
	setAttribute(Qt::WA_DeleteOnClose, true);

	// etwas kartenmaterial besorgen
	Card foo;
	foo.column[0] = 12;
	foo.column[9] = 78;
	foo.column[12][0] = true;
	foo.column[12][9] = true;

	deck = new Deck;
	deck->push_back(foo);

	for(int x = 0; x < 10; x++) {
		Card c;
		c.column[x] = x;
		deck->push_back(c);
	}

	createGraphicalEditor();
	createActions();
	createMenus();
	createToolBars();
	createStatusBar();
	createDockWindows();
	createDevices();

	// Titel einrichten und so
	connect(this, SIGNAL(fileOpened(bool)), this, SLOT(window_update_on_file_state(bool)));
	connect(this, SIGNAL(fileOpened(bool)), this, SLOT(actions_update_on_file_state(bool)));
	// und jetzt:

	notifyFileOpened();
}

void MainWindow::createGraphicalEditor() {
	graphical_editor = new CardEditor(this); // this takes ownership nevertheless
	setCentralWidget(graphical_editor);
	// signale und so:
	connect(this, SIGNAL(cardSelected(DeckIndex)), graphical_editor, SLOT(setCard(DeckIndex)));
	connect(this, SIGNAL(fileOpened(bool)), graphical_editor, SLOT(fileOpened(bool)));
	connect(this, SIGNAL(contentsChanged(DeckIndex, DeckIndex)), graphical_editor, SLOT(contentsChanged(DeckIndex, DeckIndex)));

}


bool MainWindow::maybeSave() {
	if(!deck->isModified())
		return true;
	if(deck->empty()) // oder sowas in der Art -- eine leere Lochkarte oder so
		return true;
	switch(QMessageBox::warning(this, tr("Punch Card Editor"),
		tr("The document has been modified.\nDo you want to save your changes?"),
		QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel)) {

		case QMessageBox::Save:
			return saveFile();
		case QMessageBox::Cancel:
			return false; // aborted the whole thing
		default:
			return true; // discard the file
	}
}

void MainWindow::closeDeck() {
	qDebug() << "Deleting deck.";
	delete deck;
	emit fileOpened(deck); // also false
}

void MainWindow::loadDeck(const QString& filename) {
	// load a new deck from a file. There can still be an old deck open,
	// this method will close it when everything went good
	QFile file(filename);
	FileFormat format = FileFormatFactory::createFormat(
				FileFormatFactory::autoDetectFormat(file)
			     );
	Deck* new_deck = new Deck;
	// Leser anschmeissen
	if( new_deck->read(format, file) ) {
		statusBar()->showMessage(QString(tr("Deck read in successfully")), 4000);
	} else {
		QMessageBox::critical(this, tr("Could not open file"),
				      tr("Could not open file %1 using format %2 for reading. Reason: %3")
				      .arg(filename).arg(format.getName()).arg(file.errorString())
				      );
		return;
	}
	// Alte Datei rausschmeissen
	closeDeck();
	// Neue setzen
	deck = new_deck;
	notifyFileOpened();
}

void MainWindow::notifyFileOpened() {
	if(deck) {
		emit fileOpened(deck);
		if(! deck->empty())
			emit setCard( deck->createIndex(0) );
		else {
			// naja, das Deck ist *leer*. gibt also nix.
		}
	} else
		qDebug("notifyFileOpened is supposed to be called when a file was *opened*, not closed");
}

void MainWindow::newFile() {
	// KISS: Simply open a new window. Don't touch current window.
	MainWindow* window = new MainWindow();
	window->show();
	// This would open a new file in current window.
	/*
	if(maybeSave()) {
		// alte Datei schliessen.
		// d.h. defakto: Deck loeschen, neues erstellen.
		// und dann alle darueber informieren. Sehr kompliziert.
		closeDeck();
		deck = new Deck();
		notifyFileOpened();
	}*/
}

void MainWindow::openFile() {
	if(maybeSave()) {
		Q_ASSERT(deck);
		// Oeffnen-Dialog anzeigen.
		QString new_filename = QFileDialog::getOpenFileName(this,
			tr("Open Card Deck File..."),
			QFileInfo(deck->getFileName()).absolutePath(),
			tr("Jones Emulated Card Decks (*);;Card Editor XML-Files (*.xml *.cml)"));
		if(!new_filename.isEmpty()) {
			loadDeck(new_filename);
		}
	}
}

bool MainWindow::saveFile() {
	return deck->canSave() ? deck->save() : saveFileAs();
}

bool MainWindow::saveFileAs() {
	Q_ASSERT(deck);
	// GUI anzeigen zum Speichern
	QString filename = QFileDialog::getSaveFileName(this,
		tr("Save Card Deck File..."),
		QFileInfo(deck->getFileName()).absolutePath(),
		tr("Jones Emulator Card Decks (*);;Card Editor XML-Files (*.xml *.cml)"));
	if(filename.isEmpty())
		return false;
	// Eigentlich jetzt: Nachfragen, welches Dateiformat zum Benutzen und so.
	// stattdessen jetzt mal billig:
	if(!deck->hasFormat()) {
		// Es fehlt noch ein Format, nutze einfach mal das erstbeste
		deck->setFormat( FileFormatFactory::createFormat(FileFormatFactory::availableFormats()[0]) );
	}

	// Eigene Datei benutzen, statt einfach nur deck->save() aufzurufen, um ggf. Fehler und so
	// zurueckzukriegen.
	QFile file(filename);
	if(deck->save(deck->getFormat(), file)) {
		statusBar()->showMessage(QString(tr("Deck written successfully to %1").arg(filename)), 4000);
		return true;
	} else {
		QMessageBox::critical(this, tr("Saving File failed"),
				      tr("Could not write card deck to file %1, using format %2. Reason: %3")
				      .arg(filename).arg(deck->getFormat().getName()).arg(file.errorString())
		);
		return false;
	}
}

void MainWindow::exportText() {
	// Strategie: In die text_editors-Liste schauen, wenn nur ein Editor
	// drin ist, ist der zu nutzende Codec klar definiert. Sonst fragen,
	// welchen Codec benutzen.
	// Dann abspeichern oder so... hier waere natuerlich auch interessant,
	// auf bestehende Dateinamen zurueckgreifen zu koennen...
	if(this->text_editors.count() == 0) {
		// no text editor present => ask user if he wants to open one
		if(QMessageBox::question(this, tr("No Text Editor started"),
			tr("You have not started any Text Editor component yet. "
			   "Each text editor uses his own codec to translate a punch card. "
			   "Hence you need a text editor to export a text using a codec. "
			   "Do you want to open a new text editor?"),
			QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes) == QMessageBox::Yes) {
				this->newTextEditor();
		}
	} else if(this->text_editors.count() == 1) {
		// exactly *one* text editor present
		Q_ASSERT(text_editors[0]);
		text_editors[0]->exportText();
	} else {
		// multiple editors present
		// urhm... there *should* have been some other toolbar by now... but there is not... so
		// display some sort of "selection" dialog
		qDebug() << "Select text editor...";
	}
}

void MainWindow::closeFile() {
	if(maybeSave()) {
		// Fenster schliessen. Deck schliesst damit automatisch.

		// Deck schliessen und fertig.
		// closeDeck();
	}
}

void MainWindow::quit() {
	if(maybeSave()) {
		// Deck schliesst automatisch ;-)
		qApp->quit();
	}
}

void MainWindow::closeEvent(QCloseEvent* event) {
	if(maybeSave()) {
		event->accept();
	} else
		event->ignore();
}

void MainWindow::help() {
	qDebug("Display some help viewer...");
}

void MainWindow::about() {
	QMessageBox::about(this, tr("About the Punch Card Editor"),
		tr("<big><b>Punch Card Editor</b></big><br>Development version"
		   "<br>This application is GPLv3 licensed"
		   "<br>Copyright Sven Koeppel"));
}

void MainWindow::newTextEditor() {
	Text::EditorDock* editor = new Text::EditorDock(this);
	addDockWidget(Qt::BottomDockWidgetArea, editor);
	view_menu->addAction(editor->toggleViewAction());
	text_editors.push_back(editor);

	// Signale anschalten:
	connect(editor, SIGNAL(cardSelected(DeckIndex)), this, SLOT(setCard(DeckIndex)));
	connect(this, SIGNAL(cardSelected(DeckIndex)), editor, SLOT(setCard(DeckIndex)));
	connect(this, SIGNAL(fileOpened(bool)), editor, SLOT(setVisible(bool)));
	// TODO: klaeren, wer den Texteditor irgendwann mal killt
	// -- ist diese Liste wirklich noetig? (ggf. *ja* wegen Textexport)
}

void MainWindow::nextCard() {
	setCard( getCurrentIndex() + 1);
}

void MainWindow::prevCard() {
	setCard( getCurrentIndex() - 1);
}

void MainWindow::setCard(DeckIndex i) {
	// wenn keine Datei offen und so
	if(!deck) {
		qDebug("No card open, skipping setCard broadcasting");
		return;
	}
	if(! i.assertDeck(deck)) {
		qDebug() << "Index i has wrong deck: " << i;
		return;
	}
	i.normalize();

	// neuen index setzen
	current_index = i;

	// Signal weitergeben:
	emit cardSelected(i);
}

void MainWindow::newCard() {
	// append new Card behind current card
	if(!deck) return;
	if(deck->insert(current_index)) {
		setCard( current_index+1 );
	}

	/*
	int new_card_index = current_card + 1;
	if(! deck->isValid(current_card))
		new_card_index = 0; // insert card at beginning of deck
	deck->insert( current_card + 1, new Card);
	// ob das hier geht...
	emit deck->contentsChanged(current_card, deck->count());
	setCard( current_card + 1 );
	*/
}

void MainWindow::moveCardForwards() {
	if(!deck) return;
	if(deck->move(current_index, current_index+1)) {
		setCard( current_index+1 );
	}
	/*
	//DeckIndex new_index(deck, current_index + 1);
	if(!deck->isValid(current_card) || !deck->isValid(new_card_index)) {
		qDebug() << "Invalid indices for moveCardForwards";
		return;
	}
	deck->move(current_card, new_card_index);
	setCard( new_card_index );
	*/
}

void MainWindow::moveCardBackwards() {
	if(!deck) return;
	if(deck->move(current_index, current_index-1)) {
		setCard( current_index-1 );
	}
	/*
	int new_card_index = current_card - 1;
	if(!deck->isValid(current_card) || !deck->isValid(new_card_index)) {
		qDebug() << "Invalid indices for moveCardBackwards";
		return;
	}
	deck->move(current_card, new_card_index);
	setCard( new_card_index );
	*/
}

void MainWindow::createStatusBar() {
	this->statusBar()->showMessage(tr("Welcome to the Punch Card Editor"), 1000);
}

void MainWindow::createDevices() {
	QList<QString> driver_names = Device::DriverFactory::availableDrivers();
	qDebug() << "Driver names: " << driver_names;
	for(int i = 0; i < driver_names.count(); i++) {
		Device::Dock* dock = new Device::Dock(driver_names[i], this);
		this->addDockWidget(Qt::TopDockWidgetArea, dock);
		dock->setVisible(false);
		device_docks.append(dock);
		devices_menu->addAction( dock->toggleViewAction() );
	}
}


void MainWindow::createDockWindows() {
	// Kartenuebersicht:
	navigator = new Navigator::Dock(this);
	addDockWidget(Qt::LeftDockWidgetArea, navigator);
	view_menu->addAction(navigator->toggleViewAction());
	// Signale werden natuerlich dort gemacht und so.
}
