#ifndef DECK_H
#define DECK_H

#include <QList>
#include <QSharedPointer>
#include <QFile>
#include <QUndoCommand>

namespace QPunchCard {
	class Deck;
	class DeckIndex;

	class DeckUndoCommand;
	class DeckAppendCard;
	class DeckModifyCard;
	// usw.
}

#include "qpunchcard/card.h"
#include "qpunchcard/format.h"

namespace QPunchCard {

/**
 * File uses the strategy pattern to switch between different
 * I/O formats (using the FileFormat class).
 **/
class Deck : public QObject {
	Q_OBJECT

	FileFormat format;
	QList< Card > cards;
	QString filename;
	QUndoStack undo_stack;
	friend class FileFormat;
	friend class DeckIndex;
	friend class DeckAppendCard;
	friend class DeckModifyCard;

	void init();

public:
	//typedef QList< QPointer<Card> >::iterator iterator;

	// =========== Constructors ==================
	/// Create empty
	Deck(QObject* parent = 0);
	~Deck() {};
	/// Create from file with format
	Deck(FileFormat format, const QString& filename, QObject* parent = 0);

	// =========== Loading and Saving, File Format Handling ==================
	// (All Functions are just Shorthands for other direct public method calls)
	bool save();
	bool save(const QString& filename);
	bool save(FileFormat format, const QString& filename);
	bool save(FileFormat format, QFile& file);
	bool read();
	bool read(const QString& filename);
	bool read(FileFormat format, const QString& filename);
	bool read(FileFormat format, QFile& file);


	bool canRead() const { return hasFile() && hasFormat(); }
	bool canSave() const { return canRead(); }
	bool hasFormat() const { return !format.isNull(); }
	FileFormat getFormat() const { return format; }
	void setFormat(FileFormat format) { this->format = format; }
	bool hasFile() const { return filename.isEmpty(); }
	void setFileName(const QString& filename) { this->filename = filename; }
	const QString& getFileName() const { return filename; }

	// =========== Undo Framework ==================
	bool isModified() const { return !undo_stack.isClean(); }
	/// for advanced usage. Handle with care. Object still owned by the Deck.
	QUndoStack* getUndoStack() { return &undo_stack; }
	QAction* createRedoAction(const QString& prefix = QString()) { return undo_stack.createRedoAction(this,prefix); }
	QAction* createUndoAction(const QString& prefix = QString()) { return undo_stack.createUndoAction(this,prefix); }


	// =========== Index Handling, reading access to cards ==================
	bool isValid(int i);
	DeckIndex createIndex(int i);
	int count() const { return cards.count(); }
	int length() const { return count(); }
	int size() const { return count(); }
	bool empty() const { return cards.empty(); }
	const Card& at(int i) const { return cards.at(i); }

public:
	// =========== Editing (Undo captured) ==================
	/**
	 * The main editing command interface. Your code creates a undo
	 * command and executes it by calling this method. It will push
	 * the command on the undo stack *AND* call command->redo() (by
	 * default), if call_redo is true.
	 * It won't return anything. If command is associated with no
	 * or another deck, it will be changed.
	 **/
	void run(DeckUndoCommand* command, bool call_redo = true);

	// all following functions are shorthands for run() with
	// appropriate commands.

	/// Insert one new card
	bool insert(DeckIndex after);
	/// Insert `times` empty cards after DeckIndex
	bool insertTimes(DeckIndex after, int times = 1);
	/// Move card from `from` to `to`
	bool move(DeckIndex from, DeckIndex to);
	/// Removes all items from `from` to `to`
	bool erase(DeckIndex from, DeckIndex to);
	/// Append a card
	void append(Card card = Card());
	void push_back(Card card) { append(card); }

	// Quick & dirty, no official standard
	void emitChanged(DeckIndex lower, DeckIndex upper);

signals:
	// =========== The fully official signals ==================

	// wer brauchts?
	void cardCountChanged(int newCardCount); // wer braucht das?

	// das hier ist das FAVORISIERTE FORMAT:
	void contentsChanged(DeckIndex lowerCard, DeckIndex upperCard);


	void modified(bool modified);
	//void redoAvailable(bool available);
	//void undoAvailable(bool available);
	//void undoCommandAdded();

public slots:
	void redo() { undo_stack.redo(); }
	void undo() { undo_stack.undo(); }
	//void setModified(bool modified) {}
};

// Java-style iterator
//typedef QListIterator< QPointer<Card> > DeckIterator;


/**
 * CardIndex feels like an int (normalized)
 **/
class DeckIndex {
private:
	QPointer<Deck> deck;
	int i;
public:
	// behaves like int, copy constructor and operator= will be created automatically
	DeckIndex(int i = 0) : i(i) {}
	operator int() const { return normalized(); }
	int asInt() const { return normalized(); }

	DeckIndex(Deck* d, int i) : deck(d), i(i) { if(!deck) qDebug() << "NULL DECK!"; }
	bool isValid() const { return deck ? (i >= 0 && i < deck->count()) : false; }

	// "normalisierung" = in Deck Bounds
	int normalized() const { if(!deck) return i; if(i < 0) return 0; else if(i >= deck->count()) return deck->count()-1; else return i; }
	void normalize() { i = normalized(); }
	bool canNormalize() const { return deck; }

	// special positions
	bool isTop() const { return i < 0; }
	bool isEnd() const { return i >= deck->count(); }
	bool isFirst() const { return i == 0; }
	bool isLast() const { return i == deck->count(); }

	QPointer<Deck> getDeck() const { return deck; }
	void setDeck(Deck* deck) { this->deck = deck; }
	bool assertDeck(Deck* fallback) { if(!deck) deck=fallback; return deck == fallback; }
	bool hasDeck() const { return deck; }

	// ein paar abkuerzungen
	/// @returns null wenn nicht normalisierbar (also letztlich kein Deck da ist)
	// QPointer geht nicht (mehr) weil Card jetzt kein QObject mehr ist
	const Card* getCard() const { return canNormalize() ? &(deck->cards.at(normalized())) : NULL; }
	// das wird crashen. *UNGUT*
	//const Card& getCard() const { return canNormalize() ? deck->cards.at(normalized()) : Card(); }

	DeckIndex& operator++() { i++; return *this; }
	DeckIndex& operator+=(int x) { i+= x; return *this; }
};

/**
 * The base undo command. Still abstract.
 **/
class DeckUndoCommand : public QUndoCommand {
public:
	DeckUndoCommand(Deck* deck = 0, QUndoCommand* parent = 0) : QUndoCommand(parent) {
		this->deck = deck;
	}
	//const QPointer<Deck> getDeck() const { return deck; }
protected:
	QPointer<Deck> deck;
};

/**
 * Lauter schoene UndoCommands
 **/
class DeckAppendCard : public DeckUndoCommand {
public:
	DeckAppendCard(Deck* deck, Card* card, QUndoCommand* parent = 0) : DeckUndoCommand(deck, parent) {
		this->deck = deck;
		this->card = card;
		setText(deck->tr("Adding card to deck"));
	}
	void undo() { deck->cards.removeLast(); }
	void redo() { deck->cards.push_back(*card); }
private:
	Deck* deck;
	Card* card;
};

class DeckModifyCard : public DeckUndoCommand {
public:
	DeckModifyCard(Deck* deck, DeckIndex& position, Card new_content, QUndoCommand* parent=0)
			: DeckUndoCommand(deck, parent), position(position), new_content(new_content) {
		position.assertDeck(deck);
		Q_ASSERT(this->deck);

		// alten Inhalt irgendwie sichern
		old_content = *position.getCard();
		setText(position.getDeck()->tr("Changed contents of card").arg(position));
	}
	void undo() {
		position.getDeck()->cards[position] = old_content;
		deck->emitChanged(position, position);
	}
	void redo() {
		position.getDeck()->cards[position] = new_content;
		deck->emitChanged(position, position);
	}
private:
	DeckIndex position;
	Card new_content;
	Card old_content;
};

// usw.

QDebug operator<<(QDebug dbg, const Deck &c);
QDebug operator<<(QDebug dbg, const DeckIndex &c);

}; // namespace

#endif /* DECK_H */
