#include <QDebug>

#include "jones.h"

using namespace QPunchCard;


bool JonesFileFormat::read(QFile& file, Deck* deck) const {
	qDebug() << "Jones reading";

	if(!file.isOpen()) {
		// oeffne Datei
		if(! file.open(QIODevice::ReadOnly))
			return false;
		qDebug() << "Opened file" << file.fileName();
	}

	// Now we have an open file at `file`. Use
	// QIODevice methods:

	// At first, check the file prefix (magic bytes)
	char char1, char2, char3;
	file.getChar(&char1);
	file.getChar(&char2);
	file.getChar(&char3);
	if( (char1 == 'H') && (char2 == '8') && (char3 == '0') ) {
		//this.default_col_length = 80; // hum... nice
		qDebug() << "Found good magic file prefix";
	} else if ((char1 == 'H') && (char2 == '8') && (char3 == '2')) {
		//this.default_col_length = 82;
		qDebug() << "H82 files not supported!";
		return false; // hum... bad
	} else {
		// Fehler spezifizieren...:
		// Input file not a card file
		qDebug() << "Could not find douglas jones magic bytes (H80)";
		return false;
	}

	// Process card deck
	int card_counter = 0;
	while( file.bytesAvailable() > 0 ) {
		//Card* cur_card = new Card();
		Card cur_card;
		qint64 card_in_file_pos = file.pos();
		// try to read in 120 (body) + 3 (header) bytes
		QByteArray data = file.read(123);

		if( ((data[0] & 0x80) == 0)
		 || ((data[1] & 0x80) == 0)
		 || ((data[2] & 0x80) == 0)) {
			// This card was corrupt or there were no more data.
			// Break it. Don't save the current card any more.
			qDebug() << "Finding no more card at position" << card_in_file_pos;
			break;
		} else {
			// some debug info
			qDebug() << "Processing card" << (card_counter++) << "at position" << card_in_file_pos;
		}

		/* This algorithm is based on the C algorithm of Jones */
		cur_card = jones_bytes_to_card(data, 3);


		/* push card on the card deck*/
		// TODO: Das muss irgendwie ggf. in das UNDO-System eingebaut
		// werden (hier falsche Stelle -- in Deck-Class muss das eingebaut
		// werden
		deck->push_back(cur_card);
	} // while ! eof

	file.close();
	return true;
} // jones reader

// jones writer
bool JonesFileFormat::write(QFile& file, const Deck* deck) const {
	qDebug() << "Jones writing";

	if(!file.isOpen()) {
		// oeffne Datei
		if(! file.open(QIODevice::WriteOnly)) {
			qDebug() << "Failed to open file: " << file.errorString();
			return false;
		}
	}

	int col_length = 80;

	// Write out file prefix
	qDebug() << "Beginning Jones writing";
	file.putChar('H');
	file.putChar('8');
	file.putChar(col_length==80 ? '0' : '2');

	//QVector::iterator i;

	//for(i = begin(); i != end(); i++ ) {
	for(int i=0; i < deck->count(); i++) {
		// iterate throught the Cards
		// erhmm... write a header...
		file.putChar(0x80);
		file.putChar(0x80);
		file.putChar(0x80);

		jones_card_to_bytes( deck->at(i), file );
	} // for cards

	file.close();
	qDebug() << "Jones File written.";
	return true; // done.
} // jones writer

void JonesFileFormat::jones_card_to_bytes(const Card& card, QIODevice& out_device) {
	for(int cur_col = 0; cur_col < 80; ) {
		char first, second, third;

		Column even_col = card.get(cur_col++);
		Column odd_col = card.get(cur_col++);

		int even = jones_column_to_integer( even_col );
		int odd = jones_column_to_integer( odd_col );

		qDebug() << "Even Col" << even_col << "=" << even;
		qDebug() << "Odd Col"  << odd_col << "=" << even;

		first = even >> 4;
		second = ((even & 017) << 4) | (odd >> 8);
		third = odd & 00377;

		out_device.putChar(first);
		out_device.putChar(second);
		out_device.putChar(third);
	} // while columns
}

QByteArray JonesFileFormat::jones_card_to_bytes(const Card& card) {
	QByteArray ret;
	QBuffer buf(&ret);
	jones_card_to_bytes(card, buf);
	return ret;
}

/**
 * You must make sure that the pointer goes into an array having more than
 * 120 elements or your application will crash...
 **/
Card JonesFileFormat::jones_bytes_to_card(const char* data) {
	Card card;

	/* This is just the algorithm written by Jones in C */
	int cur_byte = 0;
	for(int cur_col = 0; cur_col < 80; ) {
		/* read in 3 bytes */
		int first = data[cur_byte++];
		int second = data[cur_byte++];
		int third = data[cur_byte++];

		/* convert to 2 columns */
		int even_col = (first << 4) | (second >> 4);
		int odd_col = ((second & 0017) << 8) | third;

		/* save the columns in the card */
		// This is integer => Column conversion with
		// implicit knowledge that this followes the
		// Jones file format
		card.column[cur_col++] = jones_integer_to_column(even_col);
		card.column[cur_col++] = jones_integer_to_column(odd_col);
	}

	return card;
}

Card JonesFileFormat::jones_bytes_to_card(QByteArray data, int offset) {
	// make sure that data have right length (if not, append with zero)
	while(data.size() < 120 + offset)
		data.append('\0');

	// then it's nothing more than an alias
	return jones_bytes_to_card(data.data() + offset);
}

/**
 * Like Douglas Jones says in his punchcard proposal:
 *
 *     Top                  Bottom
 *       _ _ _ _ _ _ _ _ _ _ _ _
 *      |_|_|_|_|_|_|_|_|_|_|_|_|
 *     12 11 0 1 2 3 4 5 6 7 8 9
 *      |     |                 |
 *      |Zone |     Numeric     |
 *
 **/
int JonesFileFormat::jones_column_to_integer(const Column& col) {
	// real row names:             0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12
	static const int Mapping[] = { 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, 10, 11 };
	// where -1 is a dummy position outside the range since column[10] is also
	// a dummy position, being skipped in the loop ;-)

	int r = 0;
	//             das hier sollte im Idealfall gleich sein ;-)
	for(int i = 0; i < col.size() && i < (int)sizeof(Mapping); i++) {
		//qDebug("i=%d, col[i]=%d, Mapping[i]=%d", i, col[i], Mapping[i]);

		if(i != 10 && col.testBit(i)) // skip dummy position
			r |= (1 << Mapping[i]);
	}
	return r;
}

Column JonesFileFormat::jones_integer_to_column(int data) {
	static const int Mapping[] = { 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, 10, 11 };
	// where -1 is a dummy position outside the range since column[10] is also
	// a dummy position ;-)

	// this is just the inverse...
	Column r;
	for(int i = 0; i < r.size() && i < (int)sizeof(Mapping); i++) {
		if(i != 10)
			r[i] = ( (1 << Mapping[i]) & data != 0 );
	}
	return r;
}
