#ifndef CODEC_H
#define CODEC_H

#include <QString>
#include <QList>

namespace QPunchCard {
	class Codec;
}

// der define ist fuer cardcodes.h
#define ERROR     -5

#include "qpunchcard/card.h"
#include "qpunchcard/deck.h"

namespace QPunchCard {

/**
 * Abstracte Codec-Klasse. Implementierungen muessen toAscii/fromAscii
 * implementieren und sollten canEncode() implementieren. Fuer Namenszuordnung
 * und fortgeschrittene Erstellung ist CodecFactory zustaendig.
 **/
class Codec {
	const QString& name;

public:
	// soll const sein, weil Codec unveraenderbar *immer* das gleiche
	// Ergebnis liefern soll, nach Erzeugung. Das ist auch noetig fuer
	// CharArrayCodec.
	const char illegal;
	/// @param illegal_character Zeichen fuer nicht existente Zahlen
	Codec(char illegal_character = '~', const QString& name = QString()) : illegal(illegal_character), name(name) {}
	Codec(const Codec& other) : illegal(other.illegal), name(other.name) {}
	virtual ~Codec() {}

	virtual const QString& getName() const { return name; }

	virtual char toAscii(const Column* col) const = 0;
	virtual QString toAscii(const Card* target_card) const;
	virtual Column fromAscii(char ch) const = 0;
	void fromAscii(const QString& string, Card* target_card) const;

	virtual bool canEncode(const Column* col) const { return toAscii(col) != illegal; }
	virtual bool canEncode(char /*char*/ ) const = 0;
	bool canEncode(const QString& string) const;

	int countIllegalColumns(const Card* card) const;
	int countIllegalColumns(const Deck* deck) const;
};

/**
 * Codec-Implementierung, der mithilfe von hardgecodeten Codetabellen
 * (von Douglas Jones uebernommen) eine Umwandlung char->Column und andersrum
 * durchfuehren kann. Letzteres geht genauso perfomant dank 4kb grosser
 * inverser Tabelle, die bei Objektkonstruktion angelegt wird.
 **/
class CharArrayCodec : public Codec {
	class Data {
	public:
		const int* table;
		char inverse_table[4096];
		int ref;
	};

	Data* d;

public:
	CharArrayCodec(const int* table, char illegal = '~', const QString& name = QString());
	CharArrayCodec(const CharArrayCodec& other) : Codec(other) {
		d = other.d;
		d->ref++;
	}
	~CharArrayCodec() { if(--d->ref == 0) delete d; }
	char toAscii(const Column* col) const { return d->inverse_table[*col]; }
	Column fromAscii(char ch) const { return Column(canEncode(ch) ? d->table[ch]: 0); }

	bool canEncode(const Column* col) const { return d->inverse_table[*col] != illegal; }
	bool canEncode(char ch) const {
		bool r = (d->table[ch] != ERROR);
		if(ch < ' ' || ch > 'z') r = false;
		//qDebug("CharArrayCodec: %c is a %s character", ch, r ? "valid" : "invalid");
		return r; }
};

/**
 * Factory-Klasse, die alle bekannten Codecs zurueckgeben kann. Wesentliches
 * Merkmal ist die Faehigkeit, mit Namen (QStrings) von Codecs umgehen zu
 * koennen und Codecs aus Namen erstellen zu koennen. Ausserdem gibt es
 * Algorithmen zum Finden des bestmoeglichen Codecs zu einer Card/einem Deck,
 * die mit Codec:countIllegalColumns() arbeitet.
 **/
class CodecFactory {
	static QHash<QString, Codec*> codec_cache;
public:
	// TODO: Codec-Caching (Codec-Constructor privatisieren, dafuer hier caching
	// betreiben) weil einige codecs recht teuer zu erstellen sind (CharArrayCodec)
	static QList<QString> availableCodecs();
	static const Codec* createCodec(const QString& name, char illegal_character = '~');
	static const QString autoDetectCodec(const Card* card);
	static const QString autoDetectCodec(const Deck* deck);
};

}; // Namespace
#endif // CODEC_H
